// CollisionHandler object must be connected to GG.scene() to run!
public class CollisionHandler extends GGen {
    // Screen parameters
    true => int KEEP_OBJECTS_ON_SCREEN;
    float frustrumHeight, frustrumWidth;

    // List of collision objects
    CollisionObject collisionObjects[0];

    fun void addCollisionObject(CollisionObject obj) {
        collisionObjects << obj;
    }

    // Returns change in velocities
    1. => float slowFactor;

    fun vec3 handleCircleCollision(CollisionObject obj1, CollisionObject obj2) {
        (now > obj1.spawn_time + obj1.no_velocity_dur) && (now > obj2.spawn_time + obj2.no_velocity_dur)
            => int calculate_velocity;

        calculateCircleCollisionPositions(obj1, obj2) @=> vec3 pos[];
        @(0, 0, 0) => vec3 vel_chg;
        if (calculate_velocity) {
            calculateCircleCollisionVelocities(obj1, obj2) @=> vec3 vel[];
            vel[0] - obj1.cur_velocity => vel_chg;
            vel[0] => obj1.cur_velocity;
            vel[1] => obj2.cur_velocity;
        }
        pos[0] => obj1.cur_pos;
        pos[1] => obj2.cur_pos;

        return vel_chg;
    }

    fun void checkCollision(CollisionObject obj1, CollisionObject obj2) {
        if (obj1.radius > 0 && obj2.radius > 0) {
            if (getNorm(obj1.cur_pos - obj2.cur_pos) < obj1.radius + obj2.radius) {
                handleCircleCollision(obj1, obj2) => vec3 vel_chg;
                makeCollisionSound((obj1.cur_pos + obj2.cur_pos) / 2, vel_chg);
            }
        }
    }

    fun void checkCollisionWithBoundary(CollisionObject obj) {
        if (obj.radius > 0) {
            // Walls
            if (obj.cur_pos.x + obj.radius > frustrumWidth / 2) {
                -1 *=> obj.cur_velocity.x;   
                2 * (obj.cur_pos.x + obj.radius - frustrumWidth / 2) -=> obj.cur_pos.x;
            } else if(obj.cur_pos.x - obj.radius < -frustrumWidth / 2) {
                -1 *=> obj.cur_velocity.x;  
                2 * (obj.cur_pos.x - obj.radius + frustrumWidth / 2) -=> obj.cur_pos.x; 
            }

            // Ceiling and floor
            if (obj.cur_pos.y + obj.radius > frustrumHeight / 2) {
                -1 *=> obj.cur_velocity.y;   
                2 * (obj.cur_pos.y + obj.radius - frustrumHeight / 2) -=> obj.cur_pos.y;
            } else if(obj.cur_pos.y - obj.radius < -frustrumHeight / 2) {
                -1 *=> obj.cur_velocity.y;  
                2 * (obj.cur_pos.y - obj.radius + frustrumHeight / 2) -=> obj.cur_pos.y; 
            }
        }
    }
    
    fun void doCollisions() {
        for (0 => int i; i+1 < collisionObjects.size(); i++) {
            for (i+1 => int j; j < collisionObjects.size(); j++) {
                checkCollision(collisionObjects[i], collisionObjects[j]);               
            }
        }
        
        if (KEEP_OBJECTS_ON_SCREEN) {
            for (0 => int i; i < collisionObjects.size(); i++) {
                checkCollisionWithBoundary(collisionObjects[i]);   
            }
        }
    }

    fun void slowObjects() {
        for (0 => int i; i< collisionObjects.size(); i++) {
            slowFactor *=> collisionObjects[i].cur_velocity;
        }
    }

    fun void update(float dt) {
        doCollisions();
        slowObjects();
    }

    // Audio handler ==============================================================
    Gain gain => Gain narrativeGain => PoleZero dcblokewav => LPF l => JCRev rev => dac;
    0.99 => dcblokewav.blockZero;
    1. => narrativeGain.gain;
    0.2 => gain.gain;
    0.01 => rev.mix;
    15000 => l.freq;

    Narrative @narrative;
    GlassPluck6 g => gain;

    fun void setOutput(UGen output) {
        rev => output;
    }

    fun void setNarrative(Narrative @narrative_) {
        narrative_ @=> narrative;
        spork ~ narrativeParametersListener();
    }

    fun void narrativeParametersListener() {
        while (true) {
            narrative.collisionGain => narrativeGain.gain;
            narrative.collisionRev => rev.mix;
            narrative.collisionSlowFactor => slowFactor;
            if (narrative.followMouseToggle != -1) {
                followMouseToggle(narrative.followMouseToggle);
                -1 => narrative.followMouseToggle;
            }
            200::ms => now;
        }
    }

    int noteOn[100];
    for (0 => int i; i < noteOn.size(); i++) {
        false => noteOn[i];
    }

    fun void makeCollisionSound(vec3 pos, vec3 vel_chg) {
        Math.clampi((1.7 * getNorm(vel_chg)) $ int, 0, 40) $ int + 30 => int note;
        Math.clampi((4 * getNorm(vel_chg)) $ int, 0, 127) => int vel;
        Math.clampf(0.06 * getNorm(vel_chg) - 0.03, 0., 0.5) => float vol;
        if (vol > 0.05 && !noteOn[note]) {
            narrative.playedNote(note, vel, vol);
            spork ~ playSound(note, vel, vol);
        }
    }
    fun void playSound(int note, int vel, float vol) {
        vol => gain.gain;
        NoteOn(note, vel, vol);
        0.05::second => now;
        NoteOff(note);
    }
    fun void NoteOn( int m, int v, float vol )
    {
        true => noteOn[m];
        v * 1.0 / 128 => float velocity;
        g.noteOn( m, velocity );
    }
    fun void NoteOff( int m )
    {
        g.noteOff( m );
        false => noteOn[m];
    }

    // These functions tinker around with the CollisionObjects directly ===========
    fun void followMouseToggle(int toggle) {
        for (0 => int i; i < collisionObjects.size(); i++) {
            toggle => collisionObjects[i].mouseConnected;
            if (toggle == 0) {
                @(0, 0, 0) => collisionObjects[i].cur_accel;
            }
        }
    }

    // Math helper functions ======================================================
    fun float getNorm(vec3 vec) {
        return Math.sqrt(vec.x * vec.x + vec.y * vec.y + vec.z * vec.z);
    }

    fun vec3 getUnitVector(vec3 vec) {
        if (getNorm(vec) == 0.) {
            return vec;
        }
        return vec / getNorm(vec);
    }

    fun float dotProduct(vec3 left, vec3 right) {
        return left.x * right.x + left.y * right.y + left.z * right.z;
    }

    // Formula from: https://stackoverflow.com/a/35212639
    fun vec3[] calculateCircleCollisionVelocities(CollisionObject obj1, CollisionObject obj2) {
        vec3 vel1, vel2;

        if (obj1.immovable) {
            obj1.cur_velocity => vel1;
            obj2.cur_velocity - 2 * dotProduct(obj2.cur_velocity - obj1.cur_velocity, obj2.cur_pos - obj1.cur_pos) 
                    / dotProduct(obj2.cur_pos - obj1.cur_pos, obj2.cur_pos - obj1.cur_pos)
                    * (obj2.cur_pos - obj1.cur_pos) => vel2;
        } else {
            obj1.cur_velocity - dotProduct(obj1.cur_velocity - obj2.cur_velocity, obj1.cur_pos - obj2.cur_pos) 
                    / dotProduct(obj1.cur_pos - obj2.cur_pos, obj1.cur_pos - obj2.cur_pos)
                    * (obj1.cur_pos - obj2.cur_pos) => vel1;
            obj2.cur_velocity - dotProduct(obj2.cur_velocity - obj1.cur_velocity, obj2.cur_pos - obj1.cur_pos) 
                    / dotProduct(obj2.cur_pos - obj1.cur_pos, obj2.cur_pos - obj1.cur_pos)
                    * (obj2.cur_pos - obj1.cur_pos) => vel2;
        }
        
        vec3 vels[0]; 
        vels << vel1;
        vels << vel2;
        return vels;
    }

    // Assumption: obj1 and obj2 are circles in collision
    fun vec3[] calculateCircleCollisionPositions(CollisionObject obj1, CollisionObject obj2) {
        obj1.cur_pos - obj2.cur_pos => vec3 diff;
        obj1.radius + obj2.radius => float radius_sum;
        (radius_sum - getNorm(diff)) * getUnitVector(diff) => vec3 shift;

        vec3 pos1, pos2;

        if (obj1.immovable) {   // obj1 has infinite mass
            obj1.cur_pos => pos1;
            obj2.cur_pos - 1.02 * shift => pos2;
        } else {
            obj1.cur_pos + obj2.radius / radius_sum * shift => pos1;
            obj2.cur_pos - obj1.radius / radius_sum * shift => pos2;
        }

        vec3 pos[0]; 
        pos << pos1;
        pos << pos2;
        return pos;
    }

    // Screen dimension handler ================================================

    1.0 => float SPP;  // framebuffer to screen pixel ratio 

    // workaround for mac retina displays
    if (GG.frameWidth() != GG.windowWidth()) {
        2.0 => SPP;
    }

    fun void updateScreenDimensions() {
        while (true) {
            GG.frameWidth() / SPP => float screenWidth;
            GG.frameHeight() / SPP => float screenHeight;
            screenWidth / screenHeight => float aspect;
            GG.camera().viewSize() => frustrumHeight;
            frustrumHeight * aspect => frustrumWidth;
            GG.nextFrame() => now;
        }
    }
    spork ~ updateScreenDimensions();
}