public class Player extends CollisionObject {
    GCircle circle --> this;

    0.3 => float BASE_SIZE;
    BASE_SIZE => radius;
    circle.sca(radius * @(1, 1, 1));

    0.2 => float BASE_BRIGHTNESS;
    0.5 => float MAX_BRIGHTNESS;

    15. => float MAX_VELOCITY;
    100. => float MAX_ACCELERATION;
    6. => float MUL_ACCELERATION;
    
    1 => int color_change;

    // Follow mouse ========================================================
    Mouse @mouse;
    false => mouseConnected;
    fun void followMouse(Mouse @mouse_) {
        mouse_ @=> mouse;
        true => mouseConnected;
    }

    // Audio Listener ======================================================
    AudioListener @audioListener;
    fun void setAudioListener(AudioListener @audioListener_) {
        audioListener_ @=> audioListener;
    }
    fun float accelerationModifier() {
        return MUL_ACCELERATION * Math.clampf(audioListener.getScaledVolume(), 0.2, 5.);
    }
    fun float sizeModifier() {
        return BASE_SIZE + Math.clampf(0.07 * audioListener.getScaledVolume(), 0., 0.05);
    }
    fun float brightnessModifier() {
        return BASE_BRIGHTNESS + Math.clampf(0.3 * audioListener.getScaledVolume(), 
                                                    0., MAX_BRIGHTNESS - BASE_BRIGHTNESS);
    }

    // Physics engine ======================================================
    @(0, 0, 0) => cur_velocity;
    @(0, 0, 0) => cur_accel;
    @(0, 0, 0) => cur_pos;

    fun void accelTowards(vec3 pos) {
        pos - cur_pos => vec3 diff;
        0. => diff.z;

        diff * accelerationModifier() => cur_accel;
        capAcceleration();
    }

    fun void capVelocity() {
        Math.min(getNorm(cur_velocity), MAX_VELOCITY) * getUnitVector(cur_velocity) => cur_velocity;
    }
    fun void capAcceleration() {
        Math.min(getNorm(cur_accel), MAX_ACCELERATION) * getUnitVector(cur_accel) => cur_accel;
    }

    fun void update(float dt) {
        if (mouseConnected) {
            accelTowards( mouse.worldPos );
        }

        // dv = a dt
        cur_accel * dt +=> cur_velocity;
        capVelocity();

        // dx = v dt
        cur_velocity * dt +=> cur_pos;

        // Update visual
        setSize(sizeModifier());
        setPosition(cur_pos);
        if (color_change) {
            setColor(cur_pos, mouse.worldPos, brightnessModifier());
        }

        // Update audio
        //setVolume(cur_pos, mouse.worldPos);
        //setPitch(cur_pos, mouse.worldPos);
    }


    // Visual stuff ===============================================================
    fun void setSize(float size_) {
        size_ => radius;
        circle.sca(radius * @(1, 1, 1));
    }
    fun void setPosition(vec3 circle_pos) {
        this.pos(circle_pos);
    }
    fun void setColor(vec3 circle_pos, vec3 mouse_pos, float brightness) {
        getPolar(circle_pos - mouse_pos) => polar vec;

        @(Math.map(Math.phase(vec), -Math.PI, Math.PI, 0, 360), Math.mag(vec) / 2., brightness) => vec3 hsv;
        Color.hsv2rgb(hsv) => circle.mat().color;
    }

    // Audio stuff ================================================================
    SinOsc osc => dac;
    0. => osc.gain;
    440 => osc.freq;

    fun void setVolume(vec3 circle_pos, vec3 mouse_pos) {
        getNorm(circle_pos - mouse_pos) => float norm;
        Math.min(0.5, 2 / Math.pow(norm, 2)) => osc.gain;
    }
    fun void setPitch(vec3 circle_pos, vec3 mouse_pos) {
        getUnitVector(circle_pos - mouse_pos) => vec3 dir;
        //440 * Math.fabs(dir.x) + 660 * Math.fabs(dir.y) => osc.freq;
        880 + 440 * dir.x + 660 * dir.y => osc.freq;
    }

    // Math helper functions ======================================================
    fun float getNorm(vec3 vec) {
        return Math.sqrt(vec.x * vec.x + vec.y * vec.y + vec.z * vec.z);
    }
    fun vec3 getUnitVector(vec3 vec) {
        if (getNorm(vec) == 0.) {
            return vec;
        }
        return vec / getNorm(vec);
    }
    fun polar getPolar(vec3 vec) {
        return (vec.x + vec.y * Math.I) $ polar;
    }
}